#include "iome.h"

int imeRegister(iomeSysInfo *INFO,iomeRegistry *REG)
{
  char *Accept_System[]={
    "x",
    "tty",
    NULL
  };
  char **IME = Accept_System;
  int n = 0;
  while (n>=0&&IME[n]) {
    if (strcasecmp(INFO->system,IME[n])==0) n=-2;
    n++;
  }
  if (n>=0) return 0;
  REG->encoding      = "big5";
  REG->class         = "ime";
  REG->accept_system = Accept_System;
  return 1;
}

/*********************************************************\
  Hashing Key = ch[0]*30 + ch[1];
  We have at least two bytes in a string, right? 
  'x'+'0x0' isn't it?

  0x0      : 0
  a-z(A-Z) : 1-26
  space    : 27
  others   : 28
  not used : 29

  30x30 = 900
\*********************************************************/
int MapChar(int c)
{
  if (c==0) return 0;
  c = tolower(c);
  if ('a'<=c&&c<='z') return c-'a'+1;
  if (c==' ') return 28;
  return 29;
}

int HashingKey(char *w)
{
  return MapChar((int)w[0])*30+MapChar((int)w[1]);
}

typedef struct {
  char * english;
  int links;
  char **chinese;
} EnglishWord;

typedef struct {
  int SIZE;
  EnglishWord ** EW;
} HashingEnglish;

int SortEnglishWord(const void * A,const void *B)
{
  EnglishWord *EA = *(EnglishWord**)A;
  EnglishWord *EB = *(EnglishWord**)B;
  return strcasecmp(EA->english,EB->english);
}

typedef struct {
  int  Size;
  char KEY[256];

  HashingEnglish HE[900];

  int  ParkingLot;
  int  Ticket;
  int  NowHere;
  int  NextAvail;

  int  EWs;         /* English Words */
  EnglishWord ** EW;
  int  CWs;
  char ** CW;       /* Chinese Words */
} DICT;

EnglishWord * SpeakEnglishPlease(DICT *dict,char *english)
{
  HashingEnglish * HE;

  if (english==NULL||*english==0) return NULL;
  dict->ParkingLot = HashingKey(english);
  HE = &dict->HE[dict->ParkingLot];

  for (dict->Ticket=0;dict->Ticket<HE->SIZE;dict->Ticket++)
    if (strcasecmp(english,HE->EW[dict->Ticket]->english)==0) return HE->EW[dict->Ticket];

  dict->ParkingLot = -1;
  return NULL;
}

EnglishWord * HabloIngles(DICT *dict)
/* Well, not quite, it's espanol. */
{
  HashingEnglish * HE;

  dict->NextAvail = 0;
  dict->NowHere   = 0;
  dict->ParkingLot = HashingKey(dict->KEY);
  HE = &dict->HE[dict->ParkingLot];

  for (dict->Ticket=0;dict->Ticket<HE->SIZE;dict->Ticket++)
    if (strncasecmp(dict->KEY,HE->EW[dict->Ticket]->english,dict->Size)==0)
    return HE->EW[dict->Ticket];

  dict->ParkingLot = -1;
  return NULL;
}

int WordShowup(iomeRecord *RECORD,DICT *dict)
{
  int n,w,i;
  int tlen,xlen;
  char buf[1024];
  EnglishWord *EW;

  dict->NextAvail = 0;

  if (dict->ParkingLot<0) return 0;
  EW = dict->HE[dict->ParkingLot].EW[dict->Ticket];
  sprintf(RECORD->ListArea,"%s:",EW->english);
  
  w = EW->links - dict->NowHere;
  if (w>5) w = 5;
  for (n=0,i=dict->NowHere;n<w;n++,i++) {
    tlen = strlen(RECORD->ListArea);
    xlen = strlen(EW->chinese[i]);
    if (tlen+xlen>42) break;
    strcpy(buf,RECORD->ListArea);
    sprintf(RECORD->ListArea,"%s %d%s",buf,n+1,EW->chinese[i]);
  }
  if (dict->NowHere>0||((n+dict->NowHere)<EW->links)) { 
    strcat(RECORD->ListArea," >");
    dict->NextAvail = n+dict->NowHere;
  }
  return 1;
}

char * AppendChineseWords(DICT *dict,char *cw)
{
  char * p;
  int n;

#ifdef __
  /* Currently I disable the chinese code checking in order to load faster */
  for (n=0;n<dict->CWs;n++) if (strcmp(cw,dict->CW[n])==0) return dict->CW[n];
#endif

  dict->CW = (char**)realloc(dict->CW,sizeof(char*)*(dict->CWs+1));
  p = strdup(cw);
  dict->CW[dict->CWs] = p;
  dict->CWs++;
  return p;
}

int AppendEnglishWords(DICT *dict, EnglishWord *EW, char *cw)
{
  char * p;
  while (*cw) {
    /* Find the next chinese word start */
    while (*cw&&(*cw==' '||*cw==9)) cw++;
    if (!(*cw)||*cw==10||*cw==13) return 0;
    p = cw; /* start */
    while (*cw&&*cw!=' '&&*cw!=9&&*cw!=10&&*cw!=13) cw++;
    *cw = 0; cw++;
    /* Append New Chinese Link */
    p = AppendChineseWords(dict,p);
    EW->chinese = (char**)realloc(EW->chinese,sizeof(char*)*(EW->links+1));
    EW->chinese[EW->links]=p;
    EW->links++;
  }
  return 1;
}

int AppendNewKey(DICT *dict,char *key)
{
  char * p = key;
  char * ew;
  EnglishWord *EW;

  /* find english word */
  while (*p&&*p!='"') p++;
  if (!*p) return 0;
  p++; ew = p;
  while (*p&&*p!='"') p++;
  if (*p) { *p=0; p++; }

  /* Search for the english word */
  EW = SpeakEnglishPlease(dict,ew);
  if (EW==NULL) {
    int PL;
    /* Append One */
    dict->EW = (EnglishWord**)realloc(dict->EW,sizeof(EnglishWord*)*(dict->EWs+1));
    EW = (EnglishWord*)malloc(sizeof(EnglishWord));
    dict->EW[dict->EWs] = EW;
    memset(EW,0,sizeof(EnglishWord));
    dict->EWs++;
    EW->english=strdup(ew);
    /* Add Hashing Table */
    PL = HashingKey(ew);
    dict->HE[PL].EW = (EnglishWord**)realloc(dict->HE[PL].EW,sizeof(EnglishWord*)*(dict->HE[PL].SIZE+1));
    dict->HE[PL].EW[dict->HE[PL].SIZE] = EW;
    dict->HE[PL].SIZE++;
  }
  AppendEnglishWords(dict,EW,p);
  return 1;
}

int LoadDictionary(DICT *dict,char *filename)
{
  FILE *fp;
  char buf[1024];
  printf("Loading %s\n",filename);
  fp = fopen(filename,"rt");
  if (fp==NULL) return 0;
  while (fgets(buf,1023,fp)!=NULL) AppendNewKey(dict,buf);
  fclose(fp);
  return 1;
}

int imeInit(iomeConfig *CONF,iomeRecord *RECORD)
{
  DICT * dict;
  char * p;
  iomeConfig PAGES;
  int n;

  dict = (DICT*)malloc(sizeof(DICT));
  memset(dict,0,sizeof(DICT));

  p = iomeWhichItem(CONF,"dict.pages");
  if (p==NULL||strlen(p)==0) return 0;

  RECORD->MethodName = iomeWhichItem(CONF,"method.dict");
  RECORD->FilterName = NULL;
  RECORD->Encoding   = "big5";
  RECORD->ListArea   = (char *)malloc(1024); RECORD->ListArea[0]=0;
  RECORD->KeyArea    = (char *)malloc(64); RECORD->KeyArea[0]=0;
  RECORD->Value      = (char *)malloc(64); RECORD->Value[0]=0;
  RECORD->Owner      = dict;
  
  printf("Initialize Dictionary Method\n");
  /* Load Dictionary */
  memset(&PAGES,0,sizeof(PAGES));
  iomeLoadConfiguration(p,&PAGES);
  for (n=0;n<PAGES.size;n++)
    if (strcmp(PAGES.items[n].key,"load")==0)
      LoadDictionary(dict,PAGES.items[n].value);
  iomeFreeConfiguration(&PAGES);
  for (n=0;n<900;n++) if (dict->HE[n].SIZE>0)
    qsort(dict->HE[n].EW,dict->HE[n].SIZE,sizeof(EnglishWord*),SortEnglishWord);
  return 1;
}

int imePrepare(iomeRecord *RECORD)
{
  DICT * dict = (DICT*)RECORD->Owner;

  memset(RECORD->KeyArea,0,64);
  memset(RECORD->ListArea,0,64);
  memset(RECORD->Value,0,64);

  dict->Size=0;
  memset(dict->KEY,0,256);
  dict->ParkingLot = -1;
  dict->NextAvail  = 0;
  dict->NowHere    = 0;

  return 1;
}

int MoveUp(DICT *dict)
{
  if (dict->ParkingLot==0&&dict->Ticket==0) return 1;
  if (dict->Ticket>0) {
    dict->Ticket--;
    return 1;
  }
  do dict->ParkingLot--; while (dict->ParkingLot>0&&dict->HE[dict->ParkingLot].SIZE==0);
  if (dict->HE[dict->ParkingLot].SIZE==0) {
    do dict->ParkingLot++; while (dict->ParkingLot<900&&dict->HE[dict->ParkingLot].SIZE==0);
    dict->Ticket = 0;
  } else dict->Ticket = dict->HE[dict->ParkingLot].SIZE-1;
  return 1;
}

int MoveDown(DICT *dict)
{
  if (dict->ParkingLot==899&&
      dict->HE[dict->ParkingLot].SIZE==(dict->Ticket+1)) return 1;
  dict->Ticket++;
  if (dict->Ticket<dict->HE[dict->ParkingLot].SIZE) return 1;
  dict->Ticket=0;
  do dict->ParkingLot++;
  while (dict->ParkingLot<900&&dict->HE[dict->ParkingLot].SIZE==0);
  if (dict->ParkingLot==900) {
    /* Find Last one */
    do dict->ParkingLot--;
    while (dict->ParkingLot>0&&dict->HE[dict->ParkingLot].SIZE==0);
    dict->Ticket = dict->HE[dict->ParkingLot].SIZE-1;
    return 0;
  }
  return 1;
}

int SendOriginalEnglish(iomeRecord *RECORD,DICT *dict)
{
  if (dict->Size==0) return IME_RT_REJECT;
  dict->KEY[dict->Size]=0;
  strcpy(RECORD->ListArea,dict->KEY);
  strcpy(RECORD->Value,dict->KEY);
  memset(RECORD->KeyArea,0,64);
  dict->Size = 0;
  memset(dict->KEY,0,256);
  return IME_RT_SEND;
}

int imeFeedKey(iomeRecord *RECORD,int ch,int state)
{
  DICT *dict = (DICT*)RECORD->Owner;
  EnglishWord *EW;
  int id;

  switch (ch) {
    case IME_KEY_ESC:
      imePrepare(RECORD);
    return IME_RT_BLOCKOUT;

    case IME_KEY_BACK:
    case IME_KEY_DELETE:
      if (dict->Size==0) break;
      dict->Size--;
      dict->KEY[dict->Size]     = 0;
      RECORD->KeyArea[dict->Size] = 0;
      if (dict->Size==0) RECORD->ListArea[0]=0;
      if (HabloIngles(dict)==NULL) return IME_RT_BLOCKOUT;
      WordShowup(RECORD,dict);
    return IME_RT_BLOCKOUT;
    break;

    case '>': /* Next List  */
      if (dict->ParkingLot>=0&&dict->NextAvail) {
        int SS = dict->HE[dict->ParkingLot].EW[dict->Ticket]->links;
        dict->NowHere = (dict->NextAvail==SS) ? 0: dict->NextAvail;
        WordShowup(RECORD,dict);
        return IME_RT_BLOCKOUT;
      }
    break;

    case IME_KEY_UP: /* move up */
      if (dict->ParkingLot<0) break;
      if (!MoveUp(dict)) break;
      WordShowup(RECORD,dict);
    return IME_RT_BLOCKOUT;

    case IME_KEY_DOWN: /* move down */
      if (dict->ParkingLot<0) break;
      if (!MoveDown(dict)) break;
      WordShowup(RECORD,dict);
    return IME_RT_BLOCKOUT;

#ifdef useless
    case IME_KEY_PAGEUP:
      if (dict->ParkingLot<0) break;
      if (!PageUp(dict)) break;
      WordShowup(RECORD,dict);
    return IME_RT_BLOCKOUT;

    case IME_KEY_PAGEDOWN:
      if (dict->ParkingLot<0) break;
      if (!PageDown(dict)) break;
      WordShowup(RECORD,dict);      
    return IME_RT_BLOCKOUT;
#endif

    case '`':
    case '~': /* Select English and Send it */
      if (dict->ParkingLot<0) return IME_RT_REJECT;
      if (dict->HE[dict->ParkingLot].SIZE<=dict->Ticket) return IME_RT_REJECT;

      EW = dict->HE[dict->ParkingLot].EW[dict->Ticket];
      if (EW==NULL) return IME_RT_REJECT;
      if (EW->english==NULL) return IME_RT_REJECT;
      if (strlen(EW->english)==0) return IME_RT_REJECT;

      strcpy(RECORD->ListArea,EW->english);
      memset(RECORD->KeyArea,0,64);
      strcpy(RECORD->Value,EW->english);
      dict->Size = 0;
      memset(dict->KEY,0,256);
    return IME_RT_SEND;

    case IME_KEY_ENTER: /* Send Original English Sentence */
    return SendOriginalEnglish(RECORD,dict);

    /* Select */
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
      if (ch>=0x100) return IME_RT_REJECT;
      id = ch - '1'+ dict->NowHere;
      if (dict->ParkingLot>=0 &&id<dict->HE[dict->ParkingLot].SIZE) {
        char * cn;
        EW = dict->HE[dict->ParkingLot].EW[dict->Ticket];
        if (id<EW->links) {
          cn = EW->chinese[id];
          dict->KEY[dict->Size]=0;
          strcpy(RECORD->ListArea,cn);
          strcpy(RECORD->Value,cn);
          memset(RECORD->KeyArea,0,64);
          dict->Size = 0;
          dict->ParkingLot = -1;
          dict->NowHere    = 0;
          dict->NextAvail  = 0;
          memset(dict->KEY,0,256);

          return IME_RT_SEND;
        }
      }

    default:
      if (ch>=0x100) return IME_RT_REJECT;
      if (ch<' '||ch>0x7e) break;
      if (dict->Size==0&&ch==' ') break;
      if (!(ch>='a'&&ch<='z')&&
          !(ch>='A'&&ch<='Z')&&
          ch!=' '&&ch!='\''&&ch!='?'&&ch!=':'&&
          ch!=';'&&ch!='['&&ch!=']'&&ch!='{'&&ch!='}') return IME_RT_REJECT;

      RECORD->KeyArea[dict->Size] = ch;
      dict->KEY[dict->Size]     = ch;
      dict->Size++;
      if (HabloIngles(dict)==NULL) {
        strcpy(RECORD->ListArea,"No match!");
        return IME_RT_BLOCKOUT;
      }
      WordShowup(RECORD,dict);
      return IME_RT_BLOCKOUT;
  }
  return IME_RT_REJECT;
}

int imeListArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeKeyArea(iomeRecord *RECORD,int x,int y)
{
#ifdef XCIN_WORKS
  return SendOriginalEnglish(RECORD,RECORD->Owner);
#else
  return 0;
#endif
}

int imeFocus(iomeRecord *RECORD)
{
  return 0;
}

int imeQuit(iomeRecord *RECORD)
{
  DICT * dict = (DICT*)RECORD->Owner;
  int n;
  /* Free CWs */
  for (n=0;n<dict->CWs;n++) free(dict->CW[n]);
  free(dict->CW);
  /* Free EWs */
  for (n=0;n<dict->EWs;n++) {
    free(dict->EW[n]->english);
    free(dict->EW[n]->chinese);
    free(dict->EW[n]);
  }
  free(dict->EW);

  free(dict);
  free(RECORD->ListArea);
  free(RECORD->KeyArea);
  free(RECORD->Value);
    
  return 1;  
}

